<?php

namespace common\logic\Order\Order;

use common\helpers\GlobalParam;
use common\helpers\Price;
use common\logic\BaseService;
use common\logic\BuilderInterface;
use common\logic\Document\Document\Document;
use common\logic\Order\OrderStatusHistory\OrderStatusHistoryBuilder;
use common\logic\Order\ProductOrder\ProductOrder;
use common\logic\Order\ProductOrder\ProductOrderBuilder;
use common\logic\Producer\Producer\Producer;
use common\logic\Producer\Producer\ProducerRepository;
use common\logic\User\CreditHistory\CreditHistory;
use common\logic\User\CreditHistory\CreditHistoryBuilder;
use common\logic\User\CreditHistory\CreditHistoryRepository;
use common\logic\User\User\UserSolver;
use yii\web\NotFoundHttpException;

class OrderBuilder extends BaseService implements BuilderInterface
{
    protected UserSolver $userSolver;
    protected OrderSolver $orderSolver;
    protected CreditHistoryRepository $creditHistoryRepository;
    protected ProducerRepository $producerRepository;
    protected CreditHistoryBuilder $creditHistoryBuilder;
    protected ProductOrderBuilder $productOrderBuilder;
    protected OrderStatusHistoryBuilder $orderStatusHistoryBuilder;
    protected OrderRepository $orderRepository;

    public function __construct()
    {
        $this->userSolver = $this->loadService(UserSolver::class);
        $this->orderSolver = $this->loadService(OrderSolver::class);
        $this->creditHistoryRepository = $this->loadService(CreditHistoryRepository::class);
        $this->producerRepository = $this->loadService(ProducerRepository::class);
        $this->creditHistoryBuilder = $this->loadService(CreditHistoryBuilder::class);
        $this->productOrderBuilder = $this->loadService(ProductOrderBuilder::class);
        $this->orderStatusHistoryBuilder = $this->loadService(OrderStatusHistoryBuilder::class);
        $this->orderRepository = $this->loadService(OrderRepository::class);
    }

    public function instanciate(): Order
    {
        $order = new Order();

        return $order;
    }

    /**
     * Initialise le montant total, le montant déjà payé et le poids de la commande.
     */
    public function init(Order $order, $taxCalculationMethod = Document::TAX_CALCULATION_METHOD_DEFAULT)
    {
        $this->initAmount($order, $taxCalculationMethod);
        $this->initPaidAmount($order);

        return $this;
    }

    /**
     * Initialise le montant de la commande.
     */
    public function initAmount(Order $order, $taxCalculationMethod = Document::TAX_CALCULATION_METHOD_DEFAULT)
    {
        $order->amount = 0;
        $order->amount_with_tax = 0;
        $order->amount_vat = [];
        $order->invoice_amount = 0;
        $order->invoice_amount_with_tax = 0;
        $order->invoice_amount_vat = [];
        $order->weight = 0;

        if (isset($order->productOrder)) {
            foreach ($order->productOrder as $productOrder) {
                $this->addAmount($order, Order::AMOUNT_TOTAL, $productOrder, $taxCalculationMethod);
                $this->addAmount($order, Order::INVOICE_AMOUNT_TOTAL, $productOrder, $taxCalculationMethod);
                $this->addWeight($order, $productOrder);
            }
        }
    }

    public function addWeight(Order $order, ProductOrder $productOrder): void
    {
        if ($productOrder->unit == 'piece') {
            if (isset($productOrder->product)) {
                $order->weight += ($productOrder->quantity * $productOrder->product->weight) / 1000;
            }
        } else {
            $order->weight += $productOrder->quantity;
        }
    }

    public function addAmount(Order $order, $typeTotal, $productOrder, $taxCalculationMethod)
    {
        $fieldNameAmount = $this->getFieldNameAmount($typeTotal);
        $fieldNameAmountWithTax = $this->getFieldNameAmount($typeTotal, 'with_tax');
        $price = $productOrder->getPriceByTypeTotal($typeTotal);
        $order->$fieldNameAmount += $price * $productOrder->quantity;
        $order->$fieldNameAmountWithTax += Price::getPriceWithTax(
                $price,
                $productOrder->taxRate->value,
                $taxCalculationMethod
            ) * $productOrder->quantity;
        $order->addVat($order, $typeTotal, $price * $productOrder->quantity, $productOrder->taxRate, $taxCalculationMethod);
    }

    public function addVat($typeTotal, $priceTotalWithoutTax, $taxRate, $taxCalculationMethod)
    {
        $fieldName = $this->getFieldNameAmount($typeTotal, 'vat');

        if (!isset($this->$fieldName[$taxRate->id])) {
            $this->$fieldName[$taxRate->id] = 0;
        }

        $this->$fieldName[$taxRate->id] += Price::getVat($priceTotalWithoutTax, $taxRate->value, $taxCalculationMethod);
    }

    /**
     * Initialise le montant payé de la commande et le retourne.
     */
    public function initPaidAmount(Order $order): void
    {
        if (isset($order->creditHistory)) {
            $history = $order->creditHistory;
        } else {
            $history = $this->creditHistoryRepository->getByOrder($order);
        }

        $order->paid_amount = 0;

        if (count($history)) {
            foreach ($history as $ch) {
                if ($ch->type == CreditHistory::TYPE_PAYMENT) {
                    $order->paid_amount += $ch->amount;
                } elseif ($ch->type == CreditHistory::TYPE_REFUND) {
                    $order->paid_amount -= $ch->amount;
                }
            }
        }
    }

    public function delete(Order $order, $force = false)
    {
        // remboursement si l'utilisateur a payé pour cette commande
        $amountPaid = $this->orderSolver->getAmount(Order::AMOUNT_PAID);

        if ($amountPaid > 0.01) {
            $this->creditHistoryBuilder->create(
                CreditHistory::TYPE_REFUND,
                $amountPaid,
                GlobalParam::getCurrentProducer(),
                $order->user,
                $this->userSolver->getCurrent(),
                $order
            );
        }

        // delete
        if ($this->producerRepository->getConfig('option_behavior_cancel_order') == Producer::BEHAVIOR_DELETE_ORDER_DELETE ||
            ($this->producerRepository->getConfig('option_behavior_cancel_order') == Producer::BEHAVIOR_DELETE_ORDER_STATUS && strlen($order->date_delete))
            || $force) {

            $this->productOrderBuilder->deleteByOrder($order);

            return $order->delete();
        }
        // status 'delete'
        elseif ($this->producerRepository->getConfig('option_behavior_cancel_order') == Producer::BEHAVIOR_DELETE_ORDER_STATUS) {
            $order->date_delete = date('Y-m-d H:i:s');

            return $order->save();
        }
    }

    /**
     * Ajuste le crédit pour que la commande soit payée.
     */
    public function processCredit(Order $order): void
    {
        if ($order->id_user) {
            $paymentStatus = $this->orderSolver->getPaymentStatus($order);

            if ($paymentStatus == Order::PAYMENT_PAID) {
                return;
            } elseif ($paymentStatus == Order::PAYMENT_SURPLUS) {
                $type = CreditHistory::TYPE_REFUND;
                $amount = $this->orderSolver->getAmount($order, Order::AMOUNT_SURPLUS);
            } elseif ($paymentStatus == Order::PAYMENT_UNPAID) {
                $type = CreditHistory::TYPE_PAYMENT;
                $amount = $this->orderSolver->getAmount($order, Order::AMOUNT_REMAINING);
            }

            $this->creditHistoryBuilder->create(
                $type,
                $amount,
                GlobalParam::getCurrentProducer(),
                $order->user,
                $this->userSolver->getCurrent(),
                $order
            );
        }
    }

    // setTillerSynchronization
    public function updateTillerSynchronization(Order $order)
    {
        $paymentStatus = $order->getPaymentStatus();

        if ($paymentStatus == Order::PAYMENT_PAID) {
            $order->tiller_synchronization = 1;
        } else {
            $order->tiller_synchronization = 0;
        }

        $order->save();

        return $order;
    }

    /**
     * Changement de statut d'une commande.
     */
    public function changeOrderStatus(Order $order, string $newStatus, string $origin): void
    {
        $orderStatusArray = GlobalParam::get('orderStatus');
        $userCurrent = $this->userSolver->getCurrent();

        switch ($newStatus) {
            case 'new-order' :
                $this->orderStatusHistoryBuilder->create($order, $userCurrent, $newStatus, $origin);
                $order->status = $newStatus;
                $order->save();
                break;
            case 'waiting-paiement-on-delivery':
                if (in_array($newStatus, $orderStatusArray[$order->status]['nextStatusAllow'])) {
                    $this->orderStatusHistoryBuilder->create($order, $userCurrent, $newStatus, $origin);
                    $order->status = $newStatus;
                    $order->save();
                }
                break;
            case 'waiting-paiement-by-credit':
                if (in_array($newStatus, $orderStatusArray[$order->status]['nextStatusAllow'])) {
                    $this->orderStatusHistoryBuilder->create($order, $userCurrent, $newStatus, $origin);
                    $order->status = $newStatus;
                    $order->save();
                }
                break;
            case 'paid-by-credit':
                if (in_array($newStatus, $orderStatusArray[$order->status]['nextStatusAllow'])) {
                    $this->orderStatusHistoryBuilder->create($order, $userCurrent, $newStatus, $origin);
                    $order->status = $newStatus;
                    $order->save();
                }
                break;
            case 'waiting-delevery' :
                if (in_array($newStatus, $orderStatusArray[$order->status]['nextStatusAllow'])) {
                    $this->orderStatusHistoryBuilder->create($order, $userCurrent, $newStatus, $origin);
                    $order->status = $newStatus;
                    $order->save();
                }
                break;
            case 'delivered':
                if (in_array($newStatus, $orderStatusArray[$order->status]['nextStatusAllow'])) {
                    $this->orderStatusHistoryBuilder->create($order, $userCurrent, $newStatus, $origin);
                    $order->status = $newStatus;
                    $order->save();
                }
                break;
            case 'refunded':
                if (in_array($newStatus, $orderStatusArray[$order->status]['nextStatusAllow'])) {
                    $this->orderStatusHistoryBuilder->create($order, $userCurrent, $newStatus, $origin);
                    $order->status = $newStatus;
                    $order->save();
                }
                break;
            case 'cancel':
                if (in_array($newStatus, $orderStatusArray[$order->status]['nextStatusAllow'])) {
                    $this->orderStatusHistoryBuilder->create($order, $userCurrent, $newStatus, $origin);
                    $order->status = $newStatus;
                    $order->save();
                }
                break;
            default:
                throw new NotFoundHttpException('Statut de commande inconnu.');
        }
    }

    // initReference
    public function generateReference(Order $order)
    {
        $idProducer = GlobalParam::getCurrentProducerId();
        $producer = $this->producerRepository->getOneById($idProducer);

        if (!$order->reference && $producer->option_order_reference_type == Producer::ORDER_REFERENCE_TYPE_YEARLY) {
            $lastOrder = $this->orderRepository->getOneLastOfYear();

            if ($lastOrder && $lastOrder->reference && strlen($lastOrder->reference) > 0) {
                $pattern = '#A([0-9]+)C([0-9]+)#';
                preg_match($pattern, $lastOrder->reference, $matches, PREG_OFFSET_CAPTURE);
                $sizeNumReference = strlen($matches[2][0]);
                $numReference = ((int)$matches[2][0]) + 1;
                $numReference = str_pad($numReference, $sizeNumReference, '0', STR_PAD_LEFT);

                $order->reference = 'A' . $matches[1][0] . 'C' . $numReference;
            } else {
                $order->reference = 'A' . date('y') . 'C0001';
            }

            $order->save();
        }
    }

    // initInvoicePrices
    public function updateInvoicePrices(Order $order, $params = []): void
    {
        foreach ($order->productOrder as $productOrder) {
            if ($productOrder->product) {
                $this->productOrderBuilder->updateInvoicePrice($productOrder, $params);
            }
        }
    }

}